import pytest
from typing import cast
from dotenv import dotenv_values

from mpesa_connect import (
    App,
    Authorization,
    STKPush,
    C2B,
    B2C,
    AccountBalance,
    TransactionStatus,
)
from mpesa_connect.enums import ResponseType, TransactionType, IdentifierType

env = dotenv_values()


@pytest.fixture(scope="module")
def app() -> App:
    return App.create_sandbox(
        consumer_key=env["CONSUMER_KEY"],
        consumer_secret=env["CONSUMER_SECRET"],
    )


@pytest.fixture
def access_token(app: App) -> str:
    return cast(str, Authorization(app).generate_token().access_token)


def test_authorization(app: App) -> None:
    auth = Authorization(app)
    result = auth.generate_token()
    assert result.response.status_code == 200 and result.access_token


def test_stkpush_process_request(app: App, access_token: str) -> None:
    stk = STKPush(app, access_token=access_token)
    result = stk.process_request(
        business_short_code=cast(int, env["STKPUSH_BUSINESS_SHORT_CODE"]),
        phone_number=cast(int, env["STKPUSH_PHONE_NUMBER"]),
        amount=1,
        call_back_url=cast(str, env["STKPUSH_CALLBACK_URL"]),
        account_reference="ABC",
        transaction_desc="Pay KSH1 to ABC",
        pass_key=env["PASS_KEY"],
    )
    # print(result)
    assert result.response.status_code == 200 and result.response_code == "0"


def test_stkpush_query(app: App, access_token: str) -> None:
    stk = STKPush(app, access_token=access_token)
    result = stk.query(
        business_short_code=cast(int, env["STKPUSH_BUSINESS_SHORT_CODE"]),
        checkout_request_id=cast(str, env["STKPUSHQUERY_CHECKOUT_REQUEST_ID"]),
        pass_key=env["PASS_KEY"],
    )
    # print(result)
    assert result.response.status_code == 200 and result.response_code == "0"


def test_c2b_register_url(app: App, access_token: str) -> None:
    short_code = cast(int, env["C2B_SHORT_CODE"])
    c2b = C2B(app, access_token=access_token)
    result = c2b.register_url(
        short_code=short_code,
        validation_url=cast(str, env["C2B_VALIDATION_URL"]),
        confirmation_url=cast(str, env["C2B_CONFIRMATION_URL"]),
        response_type=ResponseType.COMPLETED,
    )
    # print(result)
    assert result.response.status_code == 200 and result.response_code == "0"


def test_c2b_simulate(app: App, access_token: str) -> None:
    c2b = C2B(app, access_token=access_token)
    result = c2b.simulate(
        short_code=cast(int, env["C2B_SHORT_CODE"]),
        command_id=TransactionType.CUSTOMER_PAY_BILL_ONLINE,
        amount=1,
        msisdn=cast(int, env["C2B_PHONE_NUMBER"]),
        bill_ref_number="BL/001",
    )
    # print(result)
    assert result.response.status_code == 200 and result.response_code == "0"


def test_b2c_payment_request(app: App, access_token: str) -> None:
    b2c = B2C(app, access_token=access_token)
    result = b2c.payment_request(
        initiator_name=cast(str, env["B2C_INITIATOR_NAME"]),
        security_credential=cast(str, env["B2C_SECURITY_CREDENTIAL"]),
        amount=1,
        command_id=TransactionType.BUSINESS_PAYMENT,
        party_a=cast(int, env["B2C_PARTY_A"]),
        party_b=cast(int, env["B2C_PARTY_B"]),
        queue_time_out_url=cast(str, env["B2C_QUEUE_TIMEOUT_URL"]),
        result_url=cast(str, env["B2C_RESULT_URL"]),
        remarks="Business payment",
        occassion="",
    )
    # print(result)
    assert result.response.status_code == 200 and result.response_code == "0"


def test_account_balance_query(app: App, access_token: str) -> None:
    ab = AccountBalance(app, access_token=access_token)
    result = ab.query(
        initiator=cast(str, env["ACCOUNT_BALANCE_INITIATOR"]),
        security_credential=cast(str, env["ACCOUNT_BALANCE_SECURITY_CREDENTIAL"]),
        identifier_type=IdentifierType.ORGANIZATION_SHORT_CODE,
        command_id=TransactionType.ACCOUNT_BALANCE,
        party_a=cast(int, env["ACCOUNT_BALANCE_PARTY_A"]),
        queue_time_out_url=cast(str, env["ACCOUNT_BALANCE_QUEUE_TIMEOUT_URL"]),
        result_url=cast(str, env["ACCOUNT_BALANCE_RESULT_URL"]),
        remarks="Account balance",
    )
    # print(result)
    assert result.response.status_code == 200 and result.response_code == "0"


def test_transaction_status_query(app: App, access_token: str) -> None:
    ts = TransactionStatus(app, access_token=access_token)
    result = ts.query(
        initiator=cast(str, env["TRANSACTION_STATUS_INITIATOR"]),
        security_credential=cast(str, env["TRANSACTION_STATUS_SECURITY_CREDENTIAL"]),
        transaction_id=cast(str, env["TRANSACTION_STATUS_TRANSACTION_ID"]),
        identifier_type=IdentifierType.ORGANIZATION_SHORT_CODE,
        command_id=TransactionType.TRANSACTION_STATUS_QUERY,
        party_a=cast(int, env["TRANSACTION_STATUS_PARTY_A"]),
        queue_time_out_url=cast(str, env["TRANSACTION_STATUS_QUEUE_TIMEOUT_URL"]),
        result_url=cast(str, env["TRANSACTION_STATUS_RESULT_URL"]),
        remarks="Transaction status",
        occassion="",
    )
    # print(result)
    assert result.response.status_code == 200 and result.response_code == "0"



import base64
import re
from typing import Union
from datetime import datetime


def base64encode(data: str) -> str:
    return base64.b64encode(data.encode("utf8")).decode("utf8")


def str_now() -> str:
    return datetime.now().strftime("%Y%m%d%H%M%S")


def snake_case(str: str) -> str:
    return re.sub(r"((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))", r"_\1", str).lower()


def generate_password(
    *,
    business_short_code: Union[int, str],
    pass_key: str,
    timestamp: str,
) -> str:
    return base64encode(str(business_short_code) + pass_key + timestamp)



from typing import Any
from enum import Enum

from .urls import DOMAIN_PRODUCTION, DOMAIN_SANDBOX


class AppDomain(Enum):
    SANDBOX = DOMAIN_SANDBOX
    PRODUCTION = DOMAIN_PRODUCTION


class App:
    def __init__(
        self,
        *,
        domain: AppDomain,
        consumer_key: str,
        consumer_secret: str,
    ) -> None:
        self.domain = domain
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret

    @property
    def base_url(self) -> str:
        return self.domain.value

    @classmethod
    def create_sandbox(cls, **kwargs: Any) -> "App":
        return cls(domain=AppDomain.SANDBOX, **kwargs)

    @classmethod
    def create_production(cls, **kwargs: Any) -> "App":
        return cls(domain=AppDomain.PRODUCTION, **kwargs)



import re
import pytest

from mpesa_connect.utils import str_now, snake_case


def test_str_now() -> None:
    assert re.match(r"^\d{4}\d{2}\d{2}\d{2}\d{2}\d{2}$", str_now()) is not None


@pytest.mark.parametrize(
    "input,output",
    [
        ("snake_case", "snake_case"),
        ("snake123", "snake123"),
        ("SnakeCase", "snake_case"),
        ("ABCSnakeCase", "abc_snake_case"),
        ("snakeCaseABC", "snake_case_abc"),
    ],
)
def test_snake_case(input: str, output: str) -> None:
    assert snake_case(input) == output



from typing import Any, Union, Optional, Tuple
from urllib.parse import urljoin

import requests

from .base import Service
from .enums import TransactionType
from .urls import PATH_STKPUSH_PROCESSREQUEST, PATH_STKPUSHQUERY_QUERY
from .utils import str_now, generate_password


class STKPush(Service):
    def process_request(
        self,
        *,
        business_short_code: int,
        amount: int,
        phone_number: int,
        call_back_url: str,
        account_reference: str,
        transaction_desc: str,
        transaction_type: Optional[Union[TransactionType, str]] = None,
        party_a: Optional[str] = None,
        party_b: Optional[str] = None,
        password: Optional[str] = None,
        timestamp: Optional[str] = None,
        pass_key: Optional[str] = None,
        access_token: Optional[str] = None,
    ) -> Any:
        if not password:
            password, timestamp = self._generate_password(
                business_short_code,
                pass_key,
                timestamp,
            )
        payload = {
            "BusinessShortCode": business_short_code,
            "Password": password,
            "Timestamp": timestamp,
            "TransactionType": transaction_type
            or TransactionType.CUSTOMER_PAY_BILL_ONLINE.value,
            "Amount": amount,
            "PartyA": party_a or phone_number,
            "PartyB": party_b or business_short_code,
            "PhoneNumber": phone_number,
            "CallBackURL": call_back_url,
            "AccountReference": account_reference,
            "TransactionDesc": transaction_desc,
        }
        response = requests.post(
            urljoin(self.app.base_url, PATH_STKPUSH_PROCESSREQUEST),
            json=payload,
            headers={
                "Authorization": f"Bearer {access_token or self.access_token}",
            },
        )
        return self._make_result(response)

    def query(
        self,
        *,
        business_short_code: int,
        checkout_request_id: str,
        password: Optional[str] = None,
        timestamp: Optional[str] = None,
        pass_key: Optional[str] = None,
        access_token: Optional[str] = None,
    ) -> Any:
        if not password:
            password, timestamp = self._generate_password(
                business_short_code,
                pass_key,
                timestamp,
            )
        payload = {
            "BusinessShortCode": business_short_code,
            "Password": password,
            "Timestamp": timestamp,
            "CheckoutRequestID": checkout_request_id,
        }
        response = requests.post(
            urljoin(self.app.base_url, PATH_STKPUSHQUERY_QUERY),
            json=payload,
            headers={
                "Authorization": f"Bearer {access_token or self.access_token}",
            },
        )
        return self._make_result(response)

    def _generate_password(
        self,
        business_short_code: int,
        pass_key: Optional[str],
        timestamp: Optional[str],
    ) -> Tuple[str, str]:
        if not timestamp:
            timestamp = str_now()
        return (
            generate_password(
                business_short_code=business_short_code,
                pass_key=pass_key or "",
                timestamp=timestamp,
            ),
            timestamp,
        )



from typing import Any
from urllib.parse import urljoin

import requests

from .base import Service
from .urls import PATH_OAUTH_GENERATE
from .utils import base64encode


class Authorization(Service):
    def generate_token(self) -> Any:
        credentials = base64encode(
            f"{self.app.consumer_key}:{self.app.consumer_secret}"
        )
        response = requests.get(
            urljoin(self.app.base_url, PATH_OAUTH_GENERATE),
            headers={"Authorization": f"Basic {credentials}"},
        )
        return self._make_result(response)



from enum import Enum


class TransactionType(Enum):
    CUSTOMER_PAY_BILL_ONLINE = "CustomerPayBillOnline"
    CUSTOMER_BUY_GOODS_ONLINE = "CustomerBuyGoodsOnline"
    BUSINESS_PAYMENT = "BusinessPayment"
    SALARY_PAYMENT = "SalaryPayment"
    PROMOTION_PAYMENT = "PromotionPayment"
    ACCOUNT_BALANCE = "AccountBalance"
    TRANSACTION_STATUS_QUERY = "TransactionStatusQuery"


class ResponseType(Enum):
    COMPLETED = "Completed"
    CANCELED = "Canceled"


class IdentifierType(Enum):
    MSISDN = 1
    TILL_NUMBER = 2
    ORGANIZATION_SHORT_CODE = 4



from typing import Any, Optional, Tuple, List, Dict
from dataclasses import make_dataclass

from requests import Response

from .app import App
from .utils import snake_case


class Service:
    def __init__(
        self,
        app: App,
        /,
        *,
        access_token: Optional[str] = None,
    ) -> None:
        self.app = app
        self.access_token = access_token

    def _make_result(self, response: Response) -> Any:
        fields: List[Tuple[str, Any]] = [("response", Response)]
        kwargs: Dict[str, Any] = {"response": response}
        try:
            for k, v in response.json().items():
                k = snake_case(k)
                fields.append((k, type(v)))
                kwargs[k] = v
        except Exception as e:
            fields.append(("error", Exception))
            kwargs["error"] = e
        return make_dataclass(f"{self.__class__.__name__}Result", fields)(**kwargs)

        # TResult = TypeVar("TResult", bound="Result")

        # @dataclass
        # class Result:
        #     response: Response

        #     @classmethod
        #     def from_response(
        #         cls: Type[TResult],
        #         response: Response,
        #         *,
        #         property_mapper: Optional[Callable[[str], str]] = None,
        #     ) -> TResult:
        #         try:
        #             data = response.json()
        #         except JSONDecodeError:
        #             return cls(response)
        #         kwargs = {}
        #         for field in fields(cls)[1:]:
        #             mapped_name = (
        #                 field.metadata.get("mapped_name")
        #                 or (callable(property_mapper) and property_mapper(field.name))
        #                 or field.name
        #             )
        #             kwargs[field.name] = data.get(mapped_name)
        # return cls(response, **kwargs)



from typing import List

from .app import App
from .authorization import Authorization
from .stk_push import STKPush
from .c2b import C2B
from .b2c import B2C
from .transaction_status import TransactionStatus
from .account_balance import AccountBalance


__all__: List[str] = [
    "App",
    "Authorization",
    "STKPush",
    "C2B",
    "B2C",
    "TransactionStatus",
    "AccountBalance",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mpesa_connect', 'test']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.28.2,<3.0.0']

setup_kwargs = {
    'name': 'mpesa-connect',
    'version': '0.1.0',
    'description': 'A wrapper library for the Daraja Mpesa API',
    'long_description': '# MPESA CONNECT\n\nA wrapper library for the Daraja Mpesa API\n\n[![Language](https://img.shields.io/badge/language-python-green.svg)](https://python.org)\n\n## Features\n\n- Authorization\n- Mpesa Express\n  - STK Push\n  - Query\n- Customer To Business (C2B)\n    - Register URL\n    - Simulate\n- Business To Customer (B2C)\n- Account Balance\n- Transaction Status\n\n## Installation\n\n    $ pip install mpesa-connect\n\n## Usage\n\n*NOTE: Before you start, make sure to go through the official Daraja Mpesa API [documentation](https://developer.safaricom.co.ke/Documentation)* \n\nCreate an app instance. \n\n```python\nfrom mpesa_connect import App\n\n# Sandbox\napp = App.create_sandbox(consumer_key=..., consumer_secret=...)\n\n# Production\napp = App.create_production(consumer_key=..., consumer_secret=...)\n```\n\nGenerate an authorization token.\n\n```python\nfrom mpesa_connect import Authorization\n\nauth = Authorization(app)\nresult = auth.generate_token()\naccess_token = result.access_token\n```\n*You can attach this token to the service instance or include it as an argument to the api methods calls*\n\n### Mpesa Express\n\n**STK Push**\n```python\nfrom mpesa_connect import STKPush\n\nstk = STKPush(app, access_token=access_token)\nresult = stk.process_request(\n    business_short_code=...,\n    phone_number=...,\n    amount=...,\n    call_back_url=...,\n    account_reference=...,\n    transaction_desc=...,\n    password=...,\n    timestamp=...,\n)\n```\n\n**Query**\n```python\nresult = stk.query(\n    business_short_code=...,\n    checkout_request_id=...,\n    password=...,\n)\n```\nYou can use the `generate_password` helper to create a password\n\n```python\nfrom mpesa_connect.utils import generate_password\n\npassword = generate_password(\n    business_short_code=....,\n    pass_key=...,\n    timestamp=...,\n)\n```\nAlternatively, you can include the `pass_key` argument in place of `password` to auto generate the password\n\n### Customer To Business (C2B) API\n\n**Register URL**\n```python\nfrom mpesa_connect import C2B\nfrom mpesa_connect.enums import ResponseType, TransactionType\n\nc2b = C2B(app, access_token=access_token)\nresult = c2b.register_url(\n    short_code=...,\n    validation_url=...,\n    confirmation_url=...,\n    response_type=ResponseType.COMPLETED,\n)\n```\n\n**Simulate**\n```python\nresult = c2b.simulate(\n    short_code=...,\n    command_id=TransactionType.CUSTOMER_PAY_BILL_ONLINE,\n    amount=...,\n    msisdn=...,\n    bill_ref_number=...,\n)\n```\n\n### Business To Customer (B2C) API\n\n```python\nfrom mpesa_connect import B2C\nfrom mpesa_connect.enums import TransactionType\n\nb2c = B2C(app, access_token=access_token)\nresult = b2c.payment_request(\n    initiator_name=...,\n    security_credential=...,\n    amount=...,\n    command_id=TransactionType.BUSINESS_PAYMENT,\n    party_a=...,\n    party_b=...,\n    queue_time_out_url=...,\n    result_url=...,\n    remarks=...,\n    occassion=...,\n)\n```\n\n### Account Balance API\n\n```python\nfrom mpesa_connect import AccountBalance\nfrom mpesa_connect.enums import TransactionType, IdentifierType\n\nab = AccountBalance(app, access_token=access_token)\nresult = ab.query(\n    initiator=...,\n    security_credential=...,\n    command_id=TransactionType.ACCOUNT_BALANCE,\n    identifier_type=IdentifierType.ORGANIZATION_SHORT_CODE,\n    party_a=...,\n    queue_time_out_url=...,\n    result_url=...,\n    remarks=...,\n)\n```\n\n### Transaction Status API\n\n```python\nfrom mpesa_connect import TransactionStatus\nfrom mpesa_connect.enums import TransactionType, IdentifierType\n\nts = TransactionStatus(app, access_token=access_token)\nresult = ts.query(\n    initiator=...,\n    security_credential=...,\n    transaction_id=...,\n    command_id=TransactionType.TRANSACTION_STATUS_QUERY,\n    identifier_type=IdentifierType.ORGANIZATION_SHORT_CODE,\n    party_a=...,\n    queue_time_out_url=...,\n    result_url=...,\n    remarks=...,\n    occassion=...,\n)\n```\n\nAll API methods return a result object with a `response` property which is a [`requests.Response`](https://requests.readthedocs.io/en/latest/api/#requests.Response) object and other various properties corresponding to the json body of the response\n\n## Running Tests\n\nInstall dependencies\n\n    $ poetry install\n\nCreate `.env` file from [.env.example](https://github.com/enwawerueli/mpesa-connect/blob/main/.env.example) then edit it to add your app credentials and test parameters\n\n    $ cp .env.example .env\n\n Run tests\n\n    $ poetry run pytest\n\n## License\n\n[MIT](https://github.com/enwawerueli/mpesa-connect/blob/main/LICENSE)\n',
    'author': 'Emz D',
    'author_email': 'seaworndrift@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/enwawerueli/mpesa-connect',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Union, Optional
from urllib.parse import urljoin

import requests

from .base import Service
from .enums import TransactionType, ResponseType
from .app import AppDomain
from .urls import PATH_C2B_REGISTERURL, PATH_C2B_SIMULATE


class C2B(Service):
    def register_url(
        self,
        *,
        short_code: int,
        validation_url: str,
        confirmation_url: str,
        response_type: Union[ResponseType, str],
        access_token: Optional[str] = None,
    ) -> Any:
        payload = {
            "ShortCode": short_code,
            "ValidationURL": validation_url,
            "ConfirmationURL": confirmation_url,
            "ResponseType": response_type.value
            if isinstance(response_type, ResponseType)
            else response_type,
        }
        response = requests.post(
            urljoin(self.app.base_url, PATH_C2B_REGISTERURL),
            json=payload,
            headers={"Authorization": f"Bearer {access_token or self.access_token}"},
        )
        return self._make_result(response)

    def simulate(
        self,
        *,
        short_code: int,
        command_id: Union[TransactionType, str],
        amount: int,
        msisdn: int,
        bill_ref_number: Optional[str] = None,
        access_token: Optional[str] = None,
    ) -> Any:
        if self.app.domain != AppDomain.SANDBOX:
            raise Exception("Simulate is available on sandbox only")
        if isinstance(command_id, TransactionType):
            command_id = command_id.value
        payload = {
            "ShortCode": short_code,
            "CommandID": command_id,
            "Amount": amount,
            "Msisdn": msisdn,
        }
        if command_id == TransactionType.CUSTOMER_PAY_BILL_ONLINE.value:
            payload["BillRefNumber"] = bill_ref_number
        response = requests.post(
            urljoin(self.app.base_url, PATH_C2B_SIMULATE),
            json=payload,
            headers={"Authorization": f"Bearer {access_token or self.access_token}"},
        )
        return self._make_result(response)



from typing import Final

DOMAIN_SANDBOX: Final[str] = "https://sandbox.safaricom.co.ke"
DOMAIN_PRODUCTION: Final[str] = "https://api.safaricom.co.ke"
PATH_OAUTH_GENERATE: Final[str] = "/oauth/v1/generate?grant_type=client_credentials"
PATH_STKPUSH_PROCESSREQUEST: Final[str] = "/mpesa/stkpush/v1/processrequest"
PATH_STKPUSHQUERY_QUERY: Final[str] = "/mpesa/stkpushquery/v1/query"
PATH_C2B_REGISTERURL: Final[str] = "/mpesa/c2b/v1/registerurl"
PATH_C2B_SIMULATE: Final[str] = "/mpesa/c2b/v1/simulate"
PATH_B2C_PAYMENTREQUEST: Final[str] = "/mpesa/b2c/v1/paymentrequest"
PATH_TRANSACTIONSTATUS_QUERY: Final[str] = "/mpesa/transactionstatus/v1/query"
PATH_ACCOUNTBALANCE_QUERY: Final[str] = "/mpesa/accountbalance/v1/query"



from typing import Any, Union, Optional
from urllib.parse import urljoin

import requests

from .base import Service
from .enums import TransactionType, IdentifierType
from .urls import PATH_TRANSACTIONSTATUS_QUERY


class TransactionStatus(Service):
    def query(
        self,
        *,
        initiator: str,
        security_credential: str,
        transaction_id: str,
        identifier_type: Union[IdentifierType, int],
        party_a: int,
        remarks: str,
        queue_time_out_url: str,
        result_url: str,
        command_id: Union[
            TransactionType, str
        ] = TransactionType.TRANSACTION_STATUS_QUERY,
        occassion: Optional[str] = None,
        access_token: Optional[str] = None,
    ) -> Any:
        payload = {
            "Initiator": initiator,
            "SecurityCredential": security_credential,
            "CommandID": command_id.value
            if isinstance(command_id, TransactionType)
            else command_id,
            "TransactionID": transaction_id,
            "PartyA": party_a,
            "IdentifierType": identifier_type.value
            if isinstance(identifier_type, IdentifierType)
            else identifier_type,
            "Remarks": remarks,
            "QueueTimeOutURL": queue_time_out_url,
            "ResultURL": result_url,
            "Occassion": occassion,
        }
        response = requests.post(
            urljoin(self.app.base_url, PATH_TRANSACTIONSTATUS_QUERY),
            json=payload,
            headers={"Authorization": f"Bearer {access_token or self.access_token}"},
        )
        return self._make_result(response)
